//---
//	gint:kernel:cpu - CPU "driver"
//---
#include <gint/hardware.h>
#include <gint/drivers.h>
#include <gint/mmu.h>
#include <gint/std/string.h>
#include <gint/cpu.h>

/* The kernel's interrupt and exception handlers' entry points */
extern void gint_exch(void);
extern void gint_tlbh(void);
extern void gint_inth_7705(void);
extern void gint_inth_7305(void);

/* Size of exception and TLB handlers */
extern char gint_exch_size;
extern char gint_tlbh_size;

/* VBR information (generated by the kernel module) */
extern void *gint_vbr;


//---
//	Context management
//---
/* CPU context */
typedef struct
{
	sr_t SR;
	uint32_t VBR;
	uint32_t CPUOPM;
} ctx_t;

/* ctx_save(): save the CPU context */
static void ctx_save(ctx_t *ctx)
{
	ctx->VBR = (uint32_t)(uintptr_t)cpu_getVBR();
	if(isSH4()) {
		ctx->CPUOPM = cpu_getCPUOPM();
		ctx->SR = cpu_getSR();
	}
}

/* ctx_restore(): Restore the CPU context */
static void ctx_restore(ctx_t *ctx)
{
	cpu_setVBR((void*)(uintptr_t)ctx->VBR);
	if(isSH4()) {
		cpu_setCPUOPM(ctx->CPUOPM);
		cpu_setSR(ctx->SR);
	}
}
/* ctx_size(): return the context size */
static size_t ctx_size(void)
{
	return (sizeof(ctx_t));
}

//---
//	Driver manipulation
//---

/* init(): Initialize the CPU */
static void init(void)
{
	/* Event handler entry points */
	uint32_t exch_size = (uintptr_t)&gint_exch_size;
	uint32_t tlbh_size = (uintptr_t)&gint_tlbh_size;
	void *inth_entry = gint_inth_7305;
	if (isSH3())
		inth_entry = gint_inth_7705;

	/* Load the event handler entry points into memory */
	memcpy(gint_vbr + 0x100, gint_exch, exch_size);
	memcpy(gint_vbr + 0x400, gint_tlbh, tlbh_size);
	memcpy(gint_vbr + 0x600, inth_entry, 64);

	/* enable some feature on SH4 */
	if(isSH4()) {
		cpu_setCPUOPM(cpu_getCPUOPM() | 0x00000008);

		/* Enable DSP mode on the CPU */
		sr_t SR = cpu_getSR();
		SR.DSP = 1;
		cpu_setSR(SR);
	}


	/* switch the VBR */
	cpu_setVBR(gint_vbr);

}

/* Due to dire space restrictions on SH3, event codes that are translated to
   SH4 are further remapped into the VBR space to eliminate gaps and save
   space. Each entry in this table represents a 32-byte block after the
   interrupt handler's entry gate. It shows the SH4 event code whose gate is
   placed on that block (some of gint's SH4 event codes are invented to host
   helper blocks).

   For instance, the 5th block after the entry gate hosts the interrupt handler
   for SH4 event 0x9e0, which is ETMU0 underflow.

   The _inth_remap table in src/core/inth.S combines the SH3-SH4 translation
   with the compact translation, hence its entry for 0xf00 (the SH3 event code
   for ETMU0 underflow) is the offset in this table where 0x9e0 is stored,
   which is 4. */
static const uint16_t sh3_vbr_map[] = {
	0x400,  /* TMU0 underflow */
	0x420,  /* TMU1 underflow */
	0x440,  /* TMU2 underflow */
	0x460,  /* (gint custom: TMU helper) */
	0x9e0,  /* ETMU0 underflow */
	0xd00,  /* ETMU4 underflow (used as helper on SH3) */
	0xd20,  /* (gint custom: ETMU helper) */
	0xd40,  /* (gint custom: ETMU helper) */
	0xaa0,  /* RTC Periodic Interrupt */
	1,	/* (Filler to maintain the gap between 0xaa0 and 0xae0) */
	0xae0,  /* (gint custom: RTC helper) */
	0
};

/* gint_inthandler(): Install interrupt handlers */
/* Yatis note:
** Due to the hypervisor hack, we assume that the VBR is already mapped.
** FIXME
*/
void *gint_inthandler(int event_code, const void *handler, size_t size)
{
	void *dest;

	/* Normalize the event code */
	if(event_code < 0x400)
		return NULL;
	event_code &= ~0x1f;

	/* On SH3, make VBR compact. Use this offset specified in the VBR map
	   above to avoid gaps */
	if(isSH3())
	{
		int index = 0;
		while(sh3_vbr_map[index])
		{
			if((int)sh3_vbr_map[index] == event_code) break;
			index++;
		}

		/* This happens if the event has not beed added to the table,
		   ie. the compact VBR scheme does not support this code */
		if(!sh3_vbr_map[index]) return NULL;

		/* Gates are placed starting at VBR + 0x200 to save space */
		dest = (void *)(gint_vbr + 0x200 + index * 0x20);
	}
	/* On SH4, just use the code as offset */
	else
	{
		/* 0x40 is the size of the entry gate */
		dest = (void *)(gint_vbr + 0x640 + (event_code - 0x400));
	}

	return memcpy(dest, handler, size);
}

//---
//	Driver definition
//---
/* create the driver object */
gint_driver_t drv_cpu = {
	.name         = "CPU",
	.init         = init,
	.ctx_size     = &ctx_size,
	.ctx_save     = (void*)&ctx_save,
	.ctx_restore  = (void*)ctx_restore,
};

/* register the driver */
GINT_DECLARE_DRIVER(0, drv_cpu);
