//---
//	gint:core:drivers - Drivers manager
//---
#include <gint/drivers.h>
#include <gint/hardware.h>
#include <gint/atomic.h>
#include <gint/std/string.h>
#include <gint/std/stdlib.h>

#include "drivers.h"

/* external driver information */
extern gint_driver_t bdrv;
extern gint_driver_t edrv;

//---
// Context interface
//---
/* drivers_context_get_size(): Return the buffctx size */
size_t drivers_context_get_size(void)
{
	size_t size = 0;
	for driver_asc(drv) {
		if (drv->ctx_size != NULL)
			size += (drv->ctx_size() + 3) & ~3;
	}
	return (size);
}

/* drivers_create_cotext(): Generate buffer with can content all driver ctx */
void *drivers_context_create(void)
{
	return (malloc(drivers_context_get_size()));

}

/* drivers_context_duplicate(): Duplicate an environment */
int drivers_context_duplicate(void **dst, void *src)
{
	if (src == NULL)
		return (-1);
	size_t ctxsz = drivers_context_get_size();
	*dst = realloc(*dst, ctxsz);
	if (*dst == NULL)
		return (-2);
	atomic_begin();
	memcpy(*dst, src, ctxsz);
	atomic_end();
	return (0);
}

/* drivers_context_destroy(): Destroy generated context */
void drivers_context_destroy(void **buffctx)
{
	if (buffctx != NULL && *buffctx != NULL)
		free(*buffctx);
	*buffctx = NULL;
}




//---
// Drivers interface
//---
/* current env drivers information */
void *kernel_env_current = NULL;

/* driver_wait(): Wait until all driver have finished */
void drivers_wait(void)
{
	for driver_asc(drv) {
		if (drv->wait != NULL)
			drv->wait();
	}
}

/* drivers_save(): save the current drivers' context */
int drivers_save(void *buffctx)
{
	if (buffctx == NULL)
		return (-1);

	/* we need to wait until all transction has been done */
	drivers_wait();

	/* save the current drivers state into the given env */
	atomic_begin();
	buffctx += drivers_context_get_size();
	for driver_dsc(drv) {
		if (drv->ctx_size != NULL)
			buffctx -= (drv->ctx_size() + 3) & ~3;
		if (drv->ctx_save != NULL)
			drv->ctx_save(buffctx);
	}
	atomic_end();
	return (0);
}

/* drivers_restore(): restore entier drivers' context */
void *drivers_restore(void *buffctx)
{
	void *tmp;

	if (buffctx == NULL)
		return (NULL);

	/* we need to wait until all transction has been done */
	drivers_wait();

	/* switch current env then restore the saved drivers' env */
	atomic_begin();
	tmp = kernel_env_current;
	kernel_env_current = buffctx;
	for driver_asc(drv) {
		if (drv->ctx_restore != NULL)
			drv->ctx_restore(buffctx);
		if (drv->ctx_size != NULL)
			buffctx += (drv->ctx_size() + 3) & ~3;
	}
	atomic_end();

	/* return the previous env */
	return (tmp);
}

/* drivers_save_and_exit(): save the current context and init the next */
void *drivers_init(void *buffctx)
{
	void *tmp;

	if (buffctx == NULL)
		return (NULL);

	/* switch the current env then initialize it */
	atomic_begin();
	tmp = kernel_env_current;
	kernel_env_current = buffctx;
	for driver_asc(drv) {
		if (isSH3() && drv->driver_sh3 != NULL)
			drv->driver_sh3();
		if (drv->init != NULL)
			drv->init();
	}
	atomic_end();

	/* return the previous env */
	return (tmp);
}

//---
// abstraction
//---

/* drivers_switch(): Switch between two world */
void *drivers_switch(void *buffctx)
{
	drivers_save(kernel_env_current);
	return (drivers_restore(buffctx));
}

/* drivers_get_current(): return the current context information */
void *drivers_get_current(void)
{
	return (kernel_env_current);
}
